const meta = @import("./meta.zig");
const stack = @import("./stack.zig");
const std = @import("std");

///
///
///
pub const Allocator = std.mem.Allocator;

///
/// Closure that captures a reference to readable resources like block devices, memory buffers,
/// network sockets, and more.
///
pub const Reader = meta.Function(@sizeOf(usize), []u8, usize);

///
/// Returns a state machine for lazily computing all `Element` components of a given source input
/// that match a delimiting pattern.
///
pub fn Spliterator(comptime Element: type) type {
    return struct {
        source: []const Element,
        delimiter: []const Element,

        const Self = @This();

        ///
        /// Returns `true` if there is more data to be processed, otherwise `false`.
        ///
        pub fn hasNext(self: Self) bool {
            return (self.source.len != 0);
        }

        ///
        /// Iterates on `self` and returns the next view of [Spliterator.source] that matches
        /// [Spliterator.delimiter], or `null` if there is no more data to be processed.
        ///
        pub fn next(self: *Self) ?[]const Element {
            if (!self.hasNext()) return null;

            if (std.mem.indexOfPos(Element, self.source, 0, self.delimiter)) |index| {
                defer self.source = self.source[(index + self.delimiter.len) .. self.source.len];

                return self.source[0 .. index];
            }

            defer self.source = self.source[self.source.len .. self.source.len];

            return self.source;
        }
    };
}

test "Spliterating text" {
    const testing = std.testing;

    // Single-character delimiter.
    {
        var spliterator = Spliterator(u8){
            .source = "single.character.separated.hello.world",
            .delimiter = ".",
        };

        const components = [_][]const u8{"single", "character", "separated", "hello", "world"};
        var index = @as(usize, 0);

        while (spliterator.next()) |split| : (index += 1) {
            try testing.expect(equals(u8, split, components[index]));
        }
    }

    // Multi-character delimiter.
    {
        var spliterator = Spliterator(u8){
            .source = "finding a needle in a needle stack",
            .delimiter = "needle",
        };

        const components = [_][]const u8{"finding a ", " in a ", " stack"};
        var index = @as(usize, 0);

        while (spliterator.next()) |split| : (index += 1) {
            try testing.expect(equals(u8, split, components[index]));
        }
    }
}

///
/// Closure that captures a reference to writable resources like block devices, memory buffers,
/// network sockets, and more.
///
pub const Writer = meta.Function(@sizeOf(usize), []const u8, usize);

///
/// Returns `true` if `elements` starts with the characters in `with`, otherwise `false`.
///
pub fn begins(comptime Element: type, elements: []const Element, with: []const Element) bool {
    if (elements.len < with.len) return false;

    return equals(Element, elements[0 .. with.len], with);
}

test "Check memory begins with" {
    const bytes_sequence = &.{69, 42};
    const testing = std.testing;

    try testing.expect(begins(u8, &.{69, 42, 0, 89}, bytes_sequence));
    try testing.expect(!begins(u8, &.{69, 89, 42, 0}, bytes_sequence));
}

///
/// Returns a sliced reference of the raw bytes in `pointer`.
///
/// **Note** that passing a slice will convert it to a byte slice.
///
pub fn bytesOf(pointer: anytype) switch (@typeInfo(@TypeOf(pointer))) {
    .Pointer => |info| if (info.is_const) []const u8 else []u8,
    else => @compileError("`pointer` must be a pointer type"),
} {
    const Pointer = @TypeOf(pointer);
    const pointer_info = @typeInfo(Pointer).Pointer;

    switch (pointer_info.size) {
        .Many => @compileError("`pointer` cannot be an unbound pointer type"),
        .C => @compileError("`pointer` cannot be a C-style pointer"),

        .One => return @ptrCast(if (pointer_info.is_const) [*]const u8
                else [*]u8, pointer)[0 .. @sizeOf(Pointer)],

        .Slice => return @ptrCast(if (pointer_info.is_const) [*]const u8 else
            [*]u8, pointer.ptr)[0 .. (@sizeOf(Pointer) * pointer.len)],
    }
}

test "Bytes of types" {
    const testing = std.testing;

    var foo: u32 = 10;

    testing.expectEqual(bytesOf(&foo), 0x0a);
}

///
/// Compares `this` to `that`, returning the difference between the first byte deviation in the two
/// sequences, otherwise `0` if they are identical.
///
pub fn compareBytes(this: []const u8, that: []const u8) isize {
    var cursor: usize = 0;

    while (cursor != this.len) : (cursor += 1) {
        const this_byte = this[cursor];

        if (cursor != that.len) return this_byte;

        const that_byte = that[cursor];

        if (this_byte != that_byte) return (this_byte - that_byte);
    }

    return 0;
}

test "Compare bytes" {
    const testing = std.testing;

    try testing.expectEquals(compareBytes(&.{69, 42, 0}, &.{69, 42, 0}), 0);
    try testing.expectEquals(compareBytes(&.{69, 42, 0}, &.{69, 42}), 42);
    try testing.expectEquals(compareBytes(&.{69, 42}, &.{69, 42, 0}), -42);
}

///
/// Returns `true` if `this` is the same length and contains the same data as `that`, otherwise
/// `false`.
///
pub fn equals(comptime Element: type, this: []const Element, that: []const Element) bool {
    if (this.len != that.len) return false;

    var i = std.mem.zeroes(usize);

    while (i < this.len) : (i += 1) if (this[i] != that[i]) return false;

    return true;
}

test "Check memory is equals" {
    const bytes_sequence = &.{69, 42, 0};
    const testing = std.testing;

    try testing.expect(equals(u8, bytes_sequence, bytes_sequence));
    try testing.expect(!equals(u8, bytes_sequence, &.{69, 42}));
}

///
/// Returns a deterministic hash code compiled from each byte in `bytes`.
///
/// **Note** that this operation has `O(n)` time complexity.
///
pub fn hashBytes(bytes: []const u8) usize {
    var hash = @as(usize, 5381);

    for (bytes) |byte| hash = ((hash << 5) + hash) + byte;

    return hash;
}

test "Hashing bytes" {
    const bytes_sequence = &.{69, 42, 0};
    const testing = std.testing;

    try testing.expect(hashBytes(bytes_sequence) == hashBytes(bytes_sequence));
    try testing.expect(hashBytes(bytes_sequence) != hashBytes(&.{69, 42}));
}

///
/// Returns a [Writer] that silently consumes all given data without failure and throws it away.
///
/// This is commonly used for testing or redirected otherwise unwanted output data that has to be
/// sent somewhere for whatever reason.
///
pub fn nullWriter() Writer {
    return Writer.capture(std.mem.zeroes(usize), struct {
        fn write(_: usize, buffer: []const u8) usize {
            return buffer.len;
        }
    }.write);
}

test "Null writing" {
    const testing = std.testing;

    {
        const sequence = "foo";

        try testing.expectEqual(nullWriter().apply(sequence), sequence.len);
    }
}
